
#pragma once

// Genode includes
#include <base/env.h>
#include <base/heap.h>

// protobuf includes
#include <google/protobuf/message.h>

// gapfruit includes
#include <tee_session/connection.h>
#include <tee/packet_allocator.h>

// stdcxx includes
#include <functional>

namespace Tee {

	using Genode::Allocator_avl;
	using Genode::Env;
	using Genode::Heap;
	using Genode::Signal_context_capability;
	using Genode::error;
	using Genode::memset;

	using Packet_alloc_failed = Tee::Session::Tx::Source::Packet_alloc_failed;

	using google::protobuf::Message;

	struct Deserialize_failed  : Genode::Exception { };
	struct Serialize_failed    : Genode::Exception { };
	struct Transmission_error  : Genode::Exception { };

	class Connection_synchronous;
}


class Tee::Connection_synchronous
{
	private:

		enum {
			PKT_SIZE = Tee::Packet_allocator::DEFAULT_PACKET_SIZE,
			BUF_SIZE = Tee::Session::QUEUE_SIZE * PKT_SIZE,
		};

		using Signal_handler = Genode::Signal_handler<Connection_synchronous>;

		Env&                       _env;
		Heap                       _heap             { _env.ram(), _env.rm() };
		Allocator_avl              _tx_block_alloc   { &_heap };
		Tee::Connection            _connection;
		char const*                _session_label;

		Genode::Io_signal_handler<Connection_synchronous> _tx_block_sigh {_env.ep(), *this, &Connection_synchronous::tx_channel_io_signal_handler};
		Genode::Blockade _tx_status;
		void tx_channel_io_signal_handler()
		{
			_tx_status.wakeup();
		}

		Genode::Io_signal_handler<Connection_synchronous> _rx_block_sigh {_env.ep(), *this, &Connection_synchronous::rx_channel_io_signal_handler};
		Genode::Blockade _rx_status;
		void rx_channel_io_signal_handler()
		{
			_rx_status.wakeup();
		}

		// methods used by the signal handlers
		void _handle_rx_data();
		void _handle_rx_ack();
		void _handle_tx_data();
		void _handle_tx_ack();

	public:

		constexpr static size_t default_buffer_size()
		{
			return BUF_SIZE;
		}

		Connection_synchronous(Genode::Env& env, char const* label,
		                       Tee::Child_connections child_connections = Tee::Child_connections::NONE)
		:
			_env           { env },
			_connection    { env, &_tx_block_alloc, BUF_SIZE, BUF_SIZE, label, child_connections },
			_session_label { label }
		{
			_connection.sigh_tx_channel(_tx_block_sigh);
			_connection.sigh_rx_channel(_rx_block_sigh);
		}

		~Connection_synchronous() = default;

		/**
		 * Blocking send of a protobuf message
		 *
		 * \param msg                 The protobuf message to send.
		 * \param sid                 The session id. Sinks need to reply using the
		 *                            received session id so the source can track
		 *                            sessions.
		 * \param serialize_partial   Wether the msg is complete or not.
		 * \result                    Tee execution result
		 *
		 * \throw Transmission_error  E.g. when there is no sink to send to.
		 * \throw Serialize_failed    Protobuf message could not be serialized.
		 * \throw Packet_alloc_failed Could not allocate packet on connection.
		 */
		Tee::Result send(const google::protobuf::MessageLite& msg,
		                 const Tee::Session_id sid,
		                 const bool serialize_partial = false);

		/**
		 * Blocking receive of a protobuf message
		 *
		 * \param msg                 The protobuf message to receive
		 * \param sid                 The session id. Sinks need to reply using the
		 *                            received session id so the source can track
		 *                            sessions.
		 * \param parse_partial       The protobuf message to receive
		 * \result                    Tee execution result
		 *
		 * \throw Deserialize_failed  Protobuf message could not be deserialized
		 */
		Tee::Result receive(google::protobuf::MessageLite& msg,
		                    Tee::Session_id& sid,
		                    const bool parse_partial = false);
};
